@@ -16,13 +16,15 @@ package pulumi
 
 import (
 	"context"
+	"errors"
+	"fmt"
 	"reflect"
 	"sort"
 	"strings"
 
-	"github.com/pkg/errors"
 	"github.com/pulumi/pulumi/sdk/v3/go/common/resource"
 	"github.com/pulumi/pulumi/sdk/v3/go/common/resource/plugin"
+	"github.com/pulumi/pulumi/sdk/v3/go/common/util/contract"
 
 	pulumirpc "github.com/pulumi/pulumi/sdk/v3/proto/go"
 	"google.golang.org/grpc"
@@ -48,26 +50,31 @@ func construct(ctx context.Context, req *pulumirpc.ConstructRequest, engineConn
 	}
 	pulumiCtx, err := NewContext(ctx, runInfo)
 	if err != nil {
-		return nil, errors.Wrap(err, "constructing run context")
+		return nil, fmt.Errorf("constructing run context: %w", err)
 	}
 
 	// Deserialize the inputs and apply appropriate dependencies.
 	inputDependencies := req.GetInputDependencies()
 	deserializedInputs, err := plugin.UnmarshalProperties(
 		req.GetInputs(),
-		plugin.MarshalOptions{KeepSecrets: true, KeepResources: true, KeepUnknowns: req.GetDryRun()},
+		plugin.MarshalOptions{
+			KeepSecrets:      true,
+			KeepResources:    true,
+			KeepUnknowns:     req.GetDryRun(),
+			KeepOutputValues: true,
+		},
 	)
 	if err != nil {
-		return nil, errors.Wrap(err, "unmarshaling inputs")
+		return nil, fmt.Errorf("unmarshaling inputs: %w", err)
 	}
 	inputs := make(map[string]interface{}, len(deserializedInputs))
 	for key, value := range deserializedInputs {
 		k := string(key)
-		var deps []Resource
+		var deps urnSet
 		if inputDeps, ok := inputDependencies[k]; ok {
-			deps = make([]Resource, len(inputDeps.GetUrns()))
-			for i, depURN := range inputDeps.GetUrns() {
-				deps[i] = pulumiCtx.newDependencyResource(URN(depURN))
+			deps = urnSet{}
+			for _, depURN := range inputDeps.GetUrns() {
+				deps.add(URN(depURN))
 			}
 		}
 
@@ -128,7 +135,7 @@ func construct(ctx context.Context, req *pulumirpc.ConstructRequest, engineConn
 	// Serialize all state properties, first by awaiting them, and then marshaling them to the requisite gRPC values.
 	resolvedProps, propertyDeps, _, err := marshalInputs(state)
 	if err != nil {
-		return nil, errors.Wrap(err, "marshaling properties")
+		return nil, fmt.Errorf("marshaling properties: %w", err)
 	}
 
 	// Marshal all properties for the RPC call.
@@ -137,7 +144,7 @@ func construct(ctx context.Context, req *pulumirpc.ConstructRequest, engineConn
 		resolvedProps,
 		plugin.MarshalOptions{KeepSecrets: true, KeepUnknowns: keepUnknowns, KeepResources: pulumiCtx.keepResources})
 	if err != nil {
-		return nil, errors.Wrap(err, "marshaling properties")
+		return nil, fmt.Errorf("marshaling properties: %w", err)
 	}
 
 	// Convert the property dependencies map for RPC and remove duplicates.
@@ -171,7 +178,7 @@ func createProviderResource(ctx *Context, ref string) (ProviderResource, error)
 	// Parse the URN and ID out of the provider reference.
 	lastSep := strings.LastIndex(ref, "::")
 	if lastSep == -1 {
-		return nil, errors.Errorf("expected '::' in provider reference %s", ref)
+		return nil, fmt.Errorf("expected '::' in provider reference %s", ref)
 	}
 	urn := ref[0:lastSep]
 	id := ref[lastSep+2:]
@@ -190,7 +197,22 @@ func createProviderResource(ctx *Context, ref string) (ProviderResource, error)
 
 type constructInput struct {
 	value resource.PropertyValue
-	deps  []Resource
+	deps  urnSet
+}
+
+func (ci constructInput) Dependencies(ctx *Context) []Resource {
+	if ci.deps == nil {
+		return nil
+	}
+	urns := ci.deps.sortedValues()
+	var result []Resource
+	if len(urns) > 0 {
+		result = make([]Resource, len(urns))
+		for i, urn := range urns {
+			result[i] = ctx.newDependencyResource(urn)
+		}
+	}
+	return result
 }
 
 // constructInputsMap returns the inputs as a Map.
@@ -202,7 +224,7 @@ func constructInputsMap(ctx *Context, inputs map[string]interface{}) (Map, error
 		known := !ci.value.ContainsUnknowns()
 		value, secret, err := unmarshalPropertyValue(ctx, ci.value)
 		if err != nil {
-			return nil, errors.Wrapf(err, "unmarshaling input %s", k)
+			return nil, fmt.Errorf("unmarshaling input %q: %w", k, err)
 		}
 
 		resultType := anyOutputType
@@ -210,13 +232,326 @@ func constructInputsMap(ctx *Context, inputs map[string]interface{}) (Map, error
 			resultType = ot.(reflect.Type)
 		}
 
-		output := ctx.newOutput(resultType, ci.deps...)
+		output := ctx.newOutput(resultType, ci.Dependencies(ctx)...)
 		output.getState().resolve(value, known, secret, nil)
 		result[k] = output
 	}
 	return result, nil
 }
 
+func gatherDeps(v resource.PropertyValue, deps urnSet) {
+	switch {
+	case v.IsSecret():
+		gatherDeps(v.SecretValue().Element, deps)
+	case v.IsComputed():
+		gatherDeps(v.Input().Element, deps)
+	case v.IsOutput():
+		for _, urn := range v.OutputValue().Dependencies {
+			deps.add(URN(urn))
+		}
+		gatherDeps(v.OutputValue().Element, deps)
+	case v.IsResourceReference():
+		deps.add(URN(v.ResourceReferenceValue().URN))
+	case v.IsArray():
+		for _, e := range v.ArrayValue() {
+			gatherDeps(e, deps)
+		}
+	case v.IsObject():
+		for _, e := range v.ObjectValue() {
+			gatherDeps(e, deps)
+		}
+	}
+}
+
+func copyInputTo(ctx *Context, v resource.PropertyValue, dest reflect.Value) error {
+	contract.Assert(dest.CanSet())
+
+	// Check for nils. The destination will be left with the zero value.
+	if v.IsNull() {
+		return nil
+	}
+
+	// Allocate storage as necessary.
+	for dest.Kind() == reflect.Ptr {
+		elem := reflect.New(dest.Type().Elem())
+		dest.Set(elem)
+		dest = elem.Elem()
+	}
+
+	switch {
+	case v.IsSecret():
+		known, element := true, v.SecretValue().Element
+		// See if it's value is unknown.
+		if element.IsComputed() {
+			known, element = false, element.Input().Element
+		}
+		// Handle this as a secret output.
+		return copyInputTo(ctx, resource.NewOutputProperty(resource.Output{
+			Element: element,
+			Known:   known,
+			Secret:  true,
+		}), dest)
+	case v.IsComputed():
+		// Handle this as an unknown output.
+		return copyInputTo(ctx, resource.MakeOutput(v.Input().Element), dest)
+	case v.IsOutput():
+		// If it's known, there aren't any dependencies, and it's not a secret, just copy the value.
+		if v.OutputValue().Known && len(v.OutputValue().Dependencies) == 0 && !v.OutputValue().Secret {
+			return copyInputTo(ctx, v.OutputValue().Element, dest)
+		}
+
+		if !dest.Type().Implements(outputType) && !dest.Type().Implements(inputType) {
+			return fmt.Errorf("expected destination type to implement %v or %v, got %v", inputType, outputType, dest.Type())
+		}
+
+		resourceDeps := make([]Resource, len(v.OutputValue().Dependencies))
+		for i, dep := range v.OutputValue().Dependencies {
+			resourceDeps[i] = ctx.newDependencyResource(URN(dep))
+		}
+
+		result, err := createOutput(ctx, dest.Type(), v.OutputValue().Element, v.OutputValue().Known,
+			v.OutputValue().Secret, resourceDeps)
+		if err != nil {
+			return err
+		}
+		dest.Set(result)
+		return nil
+	}
+
+	if dest.Type().Implements(outputType) {
+		result, err := createOutput(ctx, dest.Type(), v, true /*known*/, false /*secret*/, nil)
+		if err != nil {
+			return err
+		}
+		dest.Set(result)
+		return nil
+	}
+
+	if dest.Type().Implements(inputType) {
+		// Try to determine the input type from the interface.
+		if it, ok := inputInterfaceTypeToConcreteType.Load(dest.Type()); ok {
+			inputType := it.(reflect.Type)
+
+			for inputType.Kind() == reflect.Ptr {
+				inputType = inputType.Elem()
+			}
+
+			switch inputType.Kind() {
+			case reflect.Bool:
+				if !v.IsBool() {
+					return fmt.Errorf("expected a %v, got a %s", inputType, v.TypeString())
+				}
+				dest.Set(reflect.ValueOf(Bool(v.BoolValue())))
+				return nil
+			case reflect.Int:
+				if !v.IsNumber() {
+					return fmt.Errorf("expected an %v, got a %s", inputType, v.TypeString())
+				}
+				dest.Set(reflect.ValueOf(Int(int64(v.NumberValue()))))
+				return nil
+			case reflect.Float64:
+				if !v.IsNumber() {
+					return fmt.Errorf("expected an %v, got a %s", inputType, v.TypeString())
+				}
+				dest.Set(reflect.ValueOf(Float64(v.NumberValue())))
+				return nil
+			case reflect.String:
+				if !v.IsString() {
+					return fmt.Errorf("expected a %v, got a %s", inputType, v.TypeString())
+				}
+				dest.Set(reflect.ValueOf(String(v.StringValue())))
+				return nil
+			case reflect.Slice:
+				return copyToSlice(ctx, v, inputType, dest)
+			case reflect.Map:
+				return copyToMap(ctx, v, inputType, dest)
+			case reflect.Interface:
+				if !anyType.Implements(inputType) {
+					return fmt.Errorf("cannot unmarshal into non-empty interface type %v", inputType)
+				}
+				result, _, err := unmarshalPropertyValue(ctx, v)
+				if err != nil {
+					return err
+				}
+				dest.Set(reflect.ValueOf(result))
+				return nil
+			case reflect.Struct:
+				return copyToStruct(ctx, v, inputType, dest)
+			default:
+				panic(fmt.Sprintf("%v", inputType.Kind()))
+			}
+		}
+
+		if v.IsAsset() {
+			if !assetType.AssignableTo(dest.Type()) {
+				return fmt.Errorf("expected a %s, got an asset", dest.Type())
+			}
+			asset, _, err := unmarshalPropertyValue(ctx, v)
+			if err != nil {
+				return err
+			}
+			dest.Set(reflect.ValueOf(asset))
+			return nil
+		}
+
+		if v.IsArchive() {
+			if !archiveType.AssignableTo(dest.Type()) {
+				return fmt.Errorf("expected a %s, got an archive", dest.Type())
+			}
+			archive, _, err := unmarshalPropertyValue(ctx, v)
+			if err != nil {
+				return err
+			}
+			dest.Set(reflect.ValueOf(archive))
+			return nil
+		}
+
+		// For plain structs that implements the Input interface.
+		if dest.Type().Kind() == reflect.Struct {
+			return copyToStruct(ctx, v, dest.Type(), dest)
+		}
+
+		// Otherwise, create an output.
+		result, err := createOutput(ctx, dest.Type(), v, true /*known*/, false /*secret*/, nil)
+		if err != nil {
+			return err
+		}
+		dest.Set(result)
+		return nil
+	}
+
+	switch dest.Type().Kind() {
+	case reflect.Map:
+		return copyToMap(ctx, v, dest.Type(), dest)
+	case reflect.Slice:
+		return copyToSlice(ctx, v, dest.Type(), dest)
+	case reflect.Struct:
+		return copyToStruct(ctx, v, dest.Type(), dest)
+	}
+
+	_, err := unmarshalOutput(ctx, v, dest)
+	if err != nil {
+		return fmt.Errorf("unmarshaling value: %w", err)
+	}
+	return nil
+}
+
+func createOutput(ctx *Context, destType reflect.Type, v resource.PropertyValue, known, secret bool,
+	deps []Resource) (reflect.Value, error) {
+
+	outputType := getOutputType(destType)
+	output := ctx.newOutput(outputType, deps...)
+	outputValueDest := reflect.New(output.ElementType()).Elem()
+	_, err := unmarshalOutput(ctx, v, outputValueDest)
+	if err != nil {
+		return reflect.Value{}, fmt.Errorf("unmarshaling value: %w", err)
+	}
+	output.getState().resolve(outputValueDest.Interface(), known, secret, nil)
+	return reflect.ValueOf(output), nil
+}
+
+func getOutputType(typ reflect.Type) reflect.Type {
+	if typ.Implements(outputType) {
+		return typ
+	} else if typ.Implements(inputType) && typ.Kind() == reflect.Interface {
+		// Attempt to determine the output type by looking up the registered input type,
+		// getting the input type's element type, and then looking up the registered output
+		// type by the element type.
+		if inputStructType, found := inputInterfaceTypeToConcreteType.Load(typ); found {
+			input := reflect.New(inputStructType.(reflect.Type)).Elem().Interface().(Input)
+			elementType := input.ElementType()
+			if outputType, ok := concreteTypeToOutputType.Load(elementType); ok {
+				return outputType.(reflect.Type)
+			}
+		}
+
+		// Otherwise, look for a `To<Name>Output` method that returns an output.
+		toOutputMethodName := "To" + strings.TrimSuffix(typ.Name(), "Input") + "Output"
+		if toOutputMethod, found := typ.MethodByName(toOutputMethodName); found {
+			mt := toOutputMethod.Type
+			if mt.NumIn() == 0 && mt.NumOut() == 1 && mt.Out(0).Implements(outputType) {
+				return mt.Out(0)
+			}
+		}
+	}
+	return anyOutputType
+}
+
+func copyToSlice(ctx *Context, v resource.PropertyValue, typ reflect.Type, dest reflect.Value) error {
+	if !v.IsArray() {
+		return fmt.Errorf("expected a %v, got a %s", typ, v.TypeString())
+	}
+	arr := v.ArrayValue()
+	slice := reflect.MakeSlice(typ, len(arr), len(arr))
+	for i, e := range arr {
+		if err := copyInputTo(ctx, e, slice.Index(i)); err != nil {
+			return err
+		}
+	}
+	dest.Set(slice)
+	return nil
+}
+
+func copyToMap(ctx *Context, v resource.PropertyValue, typ reflect.Type, dest reflect.Value) error {
+	if !v.IsObject() {
+		return fmt.Errorf("expected a %v, got a %s", typ, v.TypeString())
+	}
+
+	keyType, elemType := typ.Key(), typ.Elem()
+	if keyType.Kind() != reflect.String {
+		return fmt.Errorf("map keys must be assignable from type string")
+	}
+
+	result := reflect.MakeMap(typ)
+	for k, e := range v.ObjectValue() {
+		if resource.IsInternalPropertyKey(k) {
+			continue
+		}
+		elem := reflect.New(elemType).Elem()
+		if err := copyInputTo(ctx, e, elem); err != nil {
+			return err
+		}
+
+		key := reflect.New(keyType).Elem()
+		key.SetString(string(k))
+
+		result.SetMapIndex(key, elem)
+	}
+	dest.Set(result)
+	return nil
+}
+
+func copyToStruct(ctx *Context, v resource.PropertyValue, typ reflect.Type, dest reflect.Value) error {
+	if !v.IsObject() {
+		return fmt.Errorf("expected a %v, got a %s", typ, v.TypeString())
+	}
+	result := reflect.New(typ).Elem()
+
+	obj := v.ObjectValue()
+	for i := 0; i < typ.NumField(); i++ {
+		fieldV := result.Field(i)
+		if !fieldV.CanSet() {
+			continue
+		}
+
+		tag := typ.Field(i).Tag.Get("pulumi")
+		if tag == "" {
+			continue
+		}
+
+		e, ok := obj[resource.PropertyKey(tag)]
+		if !ok {
+			continue
+		}
+
+		if err := copyInputTo(ctx, e, fieldV); err != nil {
+			return err
+		}
+	}
+	dest.Set(result)
+	return nil
+}
+
 // constructInputsCopyTo sets the inputs on the given args struct.
 func constructInputsCopyTo(ctx *Context, inputs map[string]interface{}, args interface{}) error {
 	if args == nil {
@@ -242,19 +577,22 @@ func constructInputsCopyTo(ctx *Context, inputs map[string]interface{}, args int
 				continue
 			}
 
-			handleField := func(typ reflect.Type, value resource.PropertyValue, deps []Resource) (reflect.Value, error) {
-				resultType := anyOutputType
-				if typ.Implements(outputType) {
-					resultType = typ
-				} else if typ.Implements(inputType) {
-					toOutputMethodName := "To" + strings.TrimSuffix(typ.Name(), "Input") + "Output"
-					if toOutputMethod, found := typ.MethodByName(toOutputMethodName); found {
-						mt := toOutputMethod.Type
-						if mt.NumIn() == 0 && mt.NumOut() == 1 && mt.Out(0).Implements(outputType) {
-							resultType = mt.Out(0)
-						}
-					}
+			// Find all nested dependencies.
+			deps := urnSet{}
+			gatherDeps(ci.value, deps)
+
+			// If the top-level property dependencies are equal to (or a subset of) the gathered nested
+			// dependencies, we don't necessarily need to create a top-level output for the property.
+			if deps.contains(ci.deps) {
+				if err := copyInputTo(ctx, ci.value, fieldV); err != nil {
+					return fmt.Errorf("copying input %q: %w", k, err)
 				}
+				continue
+			}
+
+			handleField := func(typ reflect.Type, value resource.PropertyValue,
+				deps []Resource) (reflect.Value, error) {
+				resultType := getOutputType(typ)
 				output := ctx.newOutput(resultType, deps...)
 				dest := reflect.New(output.ElementType()).Elem()
 				known := !ci.value.ContainsUnknowns()
@@ -266,12 +604,12 @@ func constructInputsCopyTo(ctx *Context, inputs map[string]interface{}, args int
 				return reflect.ValueOf(output), nil
 			}
 
-			isInputType := func(typ reflect.Type) bool {
+			isOutputOrInputType := func(typ reflect.Type) bool {
 				return typ.Implements(outputType) || typ.Implements(inputType)
 			}
 
-			if isInputType(field.Type) {
-				val, err := handleField(field.Type, ci.value, ci.deps)
+			if isOutputOrInputType(field.Type) {
+				val, err := handleField(field.Type, ci.value, ci.Dependencies(ctx))
 				if err != nil {
 					return err
 				}
@@ -279,12 +617,12 @@ func constructInputsCopyTo(ctx *Context, inputs map[string]interface{}, args int
 				continue
 			}
 
-			if field.Type.Kind() == reflect.Slice && isInputType(field.Type.Elem()) {
+			if field.Type.Kind() == reflect.Slice && isOutputOrInputType(field.Type.Elem()) {
 				elemType := field.Type.Elem()
 				length := len(ci.value.ArrayValue())
 				dest := reflect.MakeSlice(field.Type, length, length)
 				for i := 0; i < length; i++ {
-					val, err := handleField(elemType, ci.value.ArrayValue()[i], ci.deps)
+					val, err := handleField(elemType, ci.value.ArrayValue()[i], ci.Dependencies(ctx))
 					if err != nil {
 						return err
 					}
@@ -294,13 +632,13 @@ func constructInputsCopyTo(ctx *Context, inputs map[string]interface{}, args int
 				continue
 			}
 
-			if field.Type.Kind() == reflect.Map && isInputType(field.Type.Elem()) {
+			if field.Type.Kind() == reflect.Map && isOutputOrInputType(field.Type.Elem()) {
 				elemType := field.Type.Elem()
 				length := len(ci.value.ObjectValue())
 				dest := reflect.MakeMapWithSize(field.Type, length)
 				for k, v := range ci.value.ObjectValue() {
 					key := reflect.ValueOf(string(k))
-					val, err := handleField(elemType, v, ci.deps)
+					val, err := handleField(elemType, v, ci.Dependencies(ctx))
 					if err != nil {
 						return err
 					}
@@ -311,19 +649,17 @@ func constructInputsCopyTo(ctx *Context, inputs map[string]interface{}, args int
 			}
 
 			if len(ci.deps) > 0 {
-				return errors.Errorf(
-					"%s.%s is typed as %v but must be typed as Input or Output for input %q with dependencies",
-					typ, field.Name, field.Type, k)
+				return fmt.Errorf("copying input %q: %s.%s is typed as %v but must be a type that implements %v or "+
+					"%v for input with dependencies", k, typ, field.Name, field.Type, inputType, outputType)
 			}
 			dest := reflect.New(field.Type).Elem()
 			secret, err := unmarshalOutput(ctx, ci.value, dest)
 			if err != nil {
-				return errors.Wrapf(err, "unmarshaling input %s", k)
+				return fmt.Errorf("copying input %q: unmarshaling value: %w", k, err)
 			}
 			if secret {
-				return errors.Errorf(
-					"%s.%s is typed as %v but must be typed as Input or Output for secret input %q",
-					typ, field.Name, field.Type, k)
+				return fmt.Errorf("copying input %q: %s.%s is typed as %v but must be a type that implements %v or "+
+					"%v for secret input", k, typ, field.Name, field.Type, inputType, outputType)
 			}
 			fieldV.Set(reflect.ValueOf(dest.Interface()))
 		}
@@ -385,26 +721,31 @@ func call(ctx context.Context, req *pulumirpc.CallRequest, engineConn *grpc.Clie
 	}
 	pulumiCtx, err := NewContext(ctx, runInfo)
 	if err != nil {
-		return nil, errors.Wrap(err, "constructing run context")
+		return nil, fmt.Errorf("constructing run context: %w", err)
 	}
 
 	// Deserialize the inputs and apply appropriate dependencies.
 	argDependencies := req.GetArgDependencies()
 	deserializedArgs, err := plugin.UnmarshalProperties(
 		req.GetArgs(),
-		plugin.MarshalOptions{KeepSecrets: true, KeepResources: true, KeepUnknowns: req.GetDryRun()},
+		plugin.MarshalOptions{
+			KeepSecrets:      true,
+			KeepResources:    true,
+			KeepUnknowns:     req.GetDryRun(),
+			KeepOutputValues: true,
+		},
 	)
 	if err != nil {
-		return nil, errors.Wrap(err, "unmarshaling inputs")
+		return nil, fmt.Errorf("unmarshaling inputs: %w", err)
 	}
 	args := make(map[string]interface{}, len(deserializedArgs))
 	for key, value := range deserializedArgs {
 		k := string(key)
-		var deps []Resource
+		var deps urnSet
 		if inputDeps, ok := argDependencies[k]; ok {
-			deps = make([]Resource, len(inputDeps.GetUrns()))
-			for i, depURN := range inputDeps.GetUrns() {
-				deps[i] = pulumiCtx.newDependencyResource(URN(depURN))
+			deps = urnSet{}
+			for _, depURN := range inputDeps.GetUrns() {
+				deps.add(URN(depURN))
 			}
 		}
 
@@ -427,7 +768,7 @@ func call(ctx context.Context, req *pulumirpc.CallRequest, engineConn *grpc.Clie
 	// Serialize all result properties, first by awaiting them, and then marshaling them to the requisite gRPC values.
 	resolvedProps, propertyDeps, _, err := marshalInputs(result)
 	if err != nil {
-		return nil, errors.Wrap(err, "marshaling properties")
+		return nil, fmt.Errorf("marshaling properties: %w", err)
 	}
 
 	// Marshal all properties for the RPC call.
@@ -436,7 +777,7 @@ func call(ctx context.Context, req *pulumirpc.CallRequest, engineConn *grpc.Clie
 		resolvedProps,
 		plugin.MarshalOptions{KeepSecrets: true, KeepUnknowns: keepUnknowns, KeepResources: pulumiCtx.keepResources})
 	if err != nil {
-		return nil, errors.Wrap(err, "marshaling properties")
+		return nil, fmt.Errorf("marshaling properties: %w", err)
 	}
 
 	// Convert the property dependencies map for RPC and remove duplicates.
@@ -495,7 +836,7 @@ func callArgsSelf(ctx *Context, source map[string]interface{}) (Resource, error)
 
 	value, secret, err := unmarshalPropertyValue(ctx, ci.value)
 	if err != nil {
-		return nil, errors.Wrap(err, "unmarshaling __self__")
+		return nil, fmt.Errorf("unmarshaling __self__: %w", err)
 	}
 	if secret {
 		return nil, errors.New("__self__ is a secret")
